package br.com.caelum.vraptor.html;

import javax.annotation.PostConstruct;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import br.com.caelum.vraptor.ioc.ApplicationScoped;
import br.com.caelum.vraptor.ioc.Component;
import br.com.caelum.vraptor.ioc.Container;

/**
 * <p>
 * Factory for implementations of {@link Url}.
 * </p>
 *
 * @author lucascs
 */
@Component
@ApplicationScoped
public class UrlFactory {

	private static Container container;

	private static ThreadLocal<VRaptorControllerUrl> last = new ThreadLocal<VRaptorControllerUrl>();

	private static final Logger LOGGER = LoggerFactory.getLogger(UrlFactory.class);

	public UrlFactory(Container container) {
		UrlFactory.container = container;
	}

	/**
	 * <p>
	 * Implemented so that VRaptor creates it at the startup of the application.
	 * Does nothing, except logging.
	 * </p>
	 */
	@PostConstruct
	public void initialize() {
		LOGGER.info("Initializing UrlFactory");
	}

	/**
	 * <p>
	 * Factory method for an {@link VRaptorControllerUrl}. Creates it and sets a
	 * ThreadLocal static field with it. Intended to be used like
	 * <code>Link link = link(); to(MyController.class).method();</code>
	 * </p>
	 *
	 * @return A fresh new instance of a UrlForLink, not yet configured
	 * @see java.lang.ThreadLocal
	 * @see UrlFactory#to(Class)
	 */
	public static Url url() {
		VRaptorControllerUrl url = container.instanceFor(VRaptorControllerUrl.class);
		LOGGER.debug("Building new link; got instance " + url
				+ " from container");
		last.set(url);
		return url;
	}

	/**
	 * <p>
	 * Factory method for a {@link PlainUrl}.
	 * </p>
	 *
	 * @param url
	 *            The literal URL to be stored in the link, or an URL relative
	 *            to the current web application. For instance, if your
	 *            application is in the context "myApp" at the host
	 *            "http://my-host.com" and you want an URL to
	 *            "http://my-host.com/myApp/some/path", you can pass here
	 *            "/some/path". This behavior is similar to the <tt>url</tt> tag
	 *            in JSTL.
	 * @return A fresh new instance of a {@link PlainUrl}, ready for use
	 */
	public static Url url(String url) {
		PlainUrl plainUrl = container.instanceFor(PlainUrl.class);
		plainUrl.saveUrlTo(url);
		return plainUrl;
	}

	/**
	 * <p>
	 * Returns a previously configured {@link VRaptorControllerUrl}. Intended to be used
	 * like <code>link(to(MyController.class).method())</code> (so the ignored
	 * argument).
	 * </p>
	 *
	 * @param whatever
	 *            Ignored
	 * @return A fresh and (hopefully) configured instance of {@link VRaptorControllerUrl}
	 * @see UrlFactory#to(Class)
	 */
	public static Url url(Object whatever) {
		VRaptorControllerUrl link = last.get();
		last.set(null);
		return link;
	}

	/**
	 * <p>
	 * Configures the {@link VRaptorControllerUrl} stored in the static ThreadLocal field.
	 * If that field is not set, creates a new {@link VRaptorControllerUrl} first, calling
	 * {@link UrlFactory#url()}.
	 * </p>
	 *
	 * @param <T>
	 *            Type of the controller class
	 * @param controller
	 *            Controller class
	 * @return A proxy of the given controller class, generated by
	 *         {@link VRaptorControllerUrl#saveUrlTo(Class)}
	 * @see java.lang.ThreadLocal
	 * @see VRaptorControllerUrl#saveUrlTo(Class)
	 * @see UrlFactory#url()
	 * @see UrlFactory#url(Object)
	 */
	public static <T> T to(Class<T> controller) {
		VRaptorControllerUrl link = last.get();
		last.set(null);
		if (link == null) {
			link = (VRaptorControllerUrl) url();
		}

		return link.saveUrlTo(controller);
	}

}
